#include <STDLIB.H>
#include <MATH.H>
#include <DOS.H>
#include <CONIO.H>

#include <TIME.H>
#include <STDIO.H>

#define DEMORUN //Comment out this line to run around with the arrow keys.

#define XDIM 80
#define YDIM 100
#define BPL  XDIM*2 //Bytes per line
#define HVAL 64*YDIM/100

//Code originally from LABDEMO2.BAS
//by Ken Silverman 14-Oct-1999

//Ported to C by Frenkel Smeijers in April 2009.

char board[64*64];
int  cosTable[256];
char texture[4*64*64];

typedef struct {
  int FPosx;
  int FPosy;
  unsigned char FAng;
} TPlayerPos;

TPlayerPos demo[3070/(sizeof(int)+sizeof(int)+sizeof(char))];


void waitForKeyPress(void) {
  if (!getch()) {
    getch();
  }
}


void setTextMode(void) {
  //80x25 16 colors
  asm {
    mov ax, 0x0003;
    int 0x10;
  }
}


void exitWithErrorMessage(char* AString) {
  setTextMode();
  printf("Wolfenstein 3-D in CGA by Frenkel.\n");
  printf("First release:               5-May-2009.\n");
  printf("Second and current release: 15-Jun-2009.\n\n");

  printf("ERROR: %s\n", AString);
  waitForKeyPress();
  exit(EXIT_FAILURE);
}


int isCGA(void) {
  //Does this computer have a VGA card or an ATI EGA Wonder card?
  //If yes then this computer does not have a CGA card,
  //if no  then this computer has a CGA card.
  char x;

  asm {
    mov ax, 0x1A00;
    int 0x10;
    mov x, al;
  }
  return x != 0x1A;
}


void loadBoard(void) {
  FILE* boardFile;

  boardFile = fopen("MAP.BIN", "rb");
  if (boardFile == NULL) {
    exitWithErrorMessage("Can't find MAP.BIN!");
  }

  fread(board, 64*64, 1, boardFile);
  fclose(boardFile);
}


void loadTexture(void) {
  //This procedure loads the textures into memory.
  //Currently it loads the textures files per byte, which is slow.
  char  pixel;
  int   x, y;
  FILE* wallblue;
  FILE* wallwhit;
  FILE* wallbrow;
  FILE* wallend;

  wallblue = fopen("WALLBLUE.BMP", "rb");
  wallwhit = fopen("WALLWHIT.BMP", "rb");
  wallbrow = fopen("WALLBROW.BMP", "rb");
  wallend  = fopen("WALLEND.BMP",  "rb");
  if ((wallblue == NULL) || (wallwhit == NULL) || (wallbrow == NULL) || (wallend == NULL)) {
    exitWithErrorMessage("Can't find WALLBLUE.BMP, WALLWHIT.BMP, WALLBROW.BMP or WALLEND.BMP!");
  }

  fseek(wallblue, 54 + 1024, SEEK_SET);
  fseek(wallwhit, 54 + 1024, SEEK_SET);
  fseek(wallbrow, 54 + 1024, SEEK_SET);
  fseek(wallend,  54 + 1024, SEEK_SET);

  for (y = 0; y < 64; y++) {
    for (x = 0; x < 64; x++) {
      fread(&pixel, 1, 1, wallblue);
      texture[x * 64 + (63 - y) + 4096 * 0] = pixel;
      fread(&pixel, 1, 1, wallwhit);
      texture[x * 64 + (63 - y) + 4096 * 1] = pixel;
      fread(&pixel, 1, 1, wallbrow);
      texture[x * 64 + (63 - y) + 4096 * 2] = pixel;
      fread(&pixel, 1, 1, wallend);
      texture[x * 64 + (63 - y) + 4096 * 3] = pixel;
    }
  }

  fclose(wallblue);
  fclose(wallwhit);
  fclose(wallbrow);
  fclose(wallend);
}


void loadCosTable(void) {
  int ang;

  for (ang = 0; ang < 256; ang++) {
    cosTable[ang] = cos(ang * M_PI / 128) * 16384;
  }
}


void setScreenMode(int AWidth, int AHeight) {
  unsigned char temp;
  int x, y;

  //Set 320x200 4 color graphics mode
  //To get full screen in Windows
  asm {
    mov ax, 0x0004;
    int 0x10;
  }

  if (AWidth == 40) {
    //Set 40x25 textmode
    asm {
      mov ax, 0x0001;
      int 0x10;
    }

    //Disable blinking
    if (isCGA()) { //CGA
      outportb(0x3D8, 8);
    }
    else { //VGA
      asm {
	mov ax, 0x1003;
	xor bl, bl;
	int 0x10;
      }
    }
  }
  else if ((AWidth == 80) || (AWidth == 160)) {
    //Set 80x25 textmode
    asm {
      mov ax, 0x0003;
      int 0x10;
    }

    //Disable blinking
    if (isCGA()) { //CGA
      outportb(0x3D8, 1 + 8);
    }
    else { //VGA
      asm {
	mov ax, 0x1003;
	xor bl, bl;
	int 0x10;
      }
    }
  }

  if (AHeight == 25) {
    //do nothing
  }
  else if (AHeight == 50) {
    if (isCGA()) { //CGA
      outportb(0x3D4, 0x04); outportb(0x3D5, 62);
      outportb(0x3D4, 0x05); outportb(0x3D5, 12);
      outportb(0x3D4, 0x06); outportb(0x3D5, 50);
      outportb(0x3D4, 0x07); outportb(0x3D5, 56);
      outportb(0x3D4, 0x09); outportb(0x3D5, 3);
    }
    else { //VGA
      outportb(0x3D4, 9);
      temp = inportb(0x3D5);
      temp = temp & (32 + 64 + 128);
      temp = temp + (8 - 1);
      outportb(0x3D4, 9);
      outportb(0x3D5, temp);
    }
  }
  else if (AHeight == 100) {
    if (isCGA()) { //CGA
      outportb(0x3D4, 0x04); outportb(0x3D5, 127);
      outportb(0x3D4, 0x05); outportb(0x3D5, 6);
      outportb(0x3D4, 0x06); outportb(0x3D5, 100);
      outportb(0x3D4, 0x07); outportb(0x3D5, 112);
      outportb(0x3D4, 0x09); outportb(0x3D5, 1);
    }
    else { //VGA
      outportb(0x3D4, 9);
      temp = inportb(0x3D5);
      temp = temp & (32 + 64 + 128);
      temp = temp + (4 - 1);
      outportb(0x3D4, 9);
      outportb(0x3D5, temp);
    }
  }

  //Initialize screen buffer
  for (y = 0; y < AHeight; y++) {
    for (x = 0; x < AWidth; x++) {
      pokeb(0xB800, y * BPL + x + x + 0, 177);//character
      pokeb(0xB800, y * BPL + x + x + 1,   0);//color
    }
  }
}


void quoteAdrianCarmack(void) {
  printf("Adrian Carmack, Master of the Pixel:\n");
  printf("\"Okay, now for the CGA version . . . NOT!\"\n");
  printf("Press any key to continue...\n");
  waitForKeyPress();
}


void drawTitle(void) {
  int   i;
  FILE* title;
  char  buffer[80*2*100];

  title = fopen("TITLE.BIN", "rb");
  if (title == NULL) {
    exitWithErrorMessage("Can't find TITLE.BIN!");
  }

  fread(buffer, 80*2*100, 1, title);

  setScreenMode(80, 100);
  for (i = 0; i < 80*2*100; i++) {
    pokeb(0xB800, i, buffer[i]);
  }

  fclose(title);
}


void loadDemo(void) {
  int i;
  TPlayerPos p;
  FILE* demoFile;

  demoFile = fopen("DEMO.BIN", "rb");
  if (demoFile == NULL) {
    exitWithErrorMessage("Can't find DEMO.BIN!");
  }

  for (i = 0; i < 614; i++) {
    fread(&p, 5, 1, demoFile);
    demo[i].FPosx = p.FPosx;
    demo[i].FPosy = p.FPosy;
    demo[i].FAng  = p.FAng;
  }

  fclose(demoFile);
}


void playMusic(void) {
  while (-1) {
    sound(262);delay(240);nosound();delay(10); //c
    sound(392);delay(240);nosound();delay(10); //g
    sound(392);delay(240);nosound();delay(10); //g
    sound(392);delay(240);nosound();delay(10); //g
    sound(587);delay(240);nosound();delay(10); //d
    sound(392);delay(240);nosound();delay(10); //g
    sound(587);delay(240);nosound();delay(10); //d
    sound(392);delay(240);nosound();delay(10); //g
    sound(262);delay(240);nosound();delay(10); //c

    delay(1000);
    if (kbhit()) {
      break;
    }

    sound(262);delay(240);nosound();delay(10); //c
    sound(247);delay(240);nosound();delay(10); //b
    sound(233);delay(240);nosound();delay(10); //ais

    sound(220);delay(240);nosound();delay(10); //a
    sound(330);delay(240);nosound();delay(10); //e
    sound(330);delay(240);nosound();delay(10); //e
    sound(330);delay(240);nosound();delay(10); //e
    sound(494);delay(240);nosound();delay(10); //b
    sound(330);delay(240);nosound();delay(10); //e
    sound(494);delay(240);nosound();delay(10); //b
    sound(330);delay(240);nosound();delay(10); //e
    sound(220);delay(240);nosound();delay(10); //a

    delay(1000);
    if (kbhit()) {
      break;
    }

    sound(220);delay(240);nosound();delay(10); //a
    sound(233);delay(240);nosound();delay(10); //ais
    sound(247);delay(240);nosound();delay(10); //b
  }
  waitForKeyPress();
}


void showStatistics(double ATime, double AFPS) {
  setTextMode();

  printf("Wolfenstein 3-D in CGA by Frenkel.\n");
  printf("First release:               5-May-2009.\n");
  printf("Second and current release: 15-Jun-2009.\n\n");
  printf("Happy 40th birthday, Adrian Carmack!\n");
  printf("Happy 17th birthday, Wolfenstein 3-D!\n\n");

  printf("Statistics:\n");
  printf("Graphics card: %s\n", (isCGA() ? "CGA" : "VGA"));
  printf("Resolution:    %dx%d\n", XDIM, YDIM);
  printf("Time:          %f seconds\n", ATime);
  printf("FPS:           %f frames per second\n", AFPS);
  waitForKeyPress();
}


int main(void) {
  int x, y;
  int playerposx;
  int playerposy;
  unsigned char playerang;
  int xscan, yscan;
  int xscaninit, yscaninit;
  int xtemp, ytemp;
  int xtempinit, ytempinit;
  int incx, incy;
  int xdir, ydir;
  int hx, hy;
  int sy1, sy2;
  int texx, texincy;//texy, texoffset;
  clock_t start, end;
  int cosang, sinang;
  int vx, vy;
  int column;
#ifdef DEMORUN
  int frame;
#else
  int quit;
  long frames;
#endif

  quoteAdrianCarmack();
  drawTitle();
  loadBoard();
  loadTexture();
  loadCosTable();
#ifdef DEMORUN
  loadDemo();
#endif
  playMusic();

  for (y = 0; y < 64; y++) {
    for (x = 0; x < 64; x++) {
      if (board[x+y*64] == 'p') {
	board[x+y*64] = 0;
	playerposx = x * 64 + 32;
	playerposy = y * 64 + 32;
      }
    }
  }
  playerang = 0;


  setScreenMode(XDIM, YDIM);

  start = clock();
#ifndef DEMORUN
  frames = 0;
  quit   = 0;
  while (quit == 0) {
#else
  for (frame = 0; frame < 614; frame++) {
    playerposx = demo[frame].FPosx;
    playerposy = demo[frame].FPosy;
    playerang  = demo[frame].FAng;
#endif

    cosang = cosTable[playerang];
    sinang = cosTable[playerang-64];
    vx = cosang + sinang;// - (sinang / XDIM);
    vy = sinang - cosang;// + (sinang / XDIM);

    xscaninit = playerposx >> 6;
    xtempinit = playerposx & 63;
    yscaninit = playerposy >> 6;
    ytempinit = playerposy & 63;
    for (column = 1; column < XDIM*2; column+=2) {
      //Raytrace in 2D to see what block is hit and where it gets hit
      xscan = xscaninit;
      yscan = yscaninit;

      //incx = abs(vx);
      //xdir = sgn(vx);
      asm {
	mov ax, vx;                   //ax = vx
	cwd;                          //if vx < 0 then dx = -1 else dx = 0
	xor ax, dx;                   //ax = vx XOR dx
	//sub ax, dx;                   //ax = (vx XOR dx) - dx
	mov incx, ax;                 //incx = abs(vx)
	or  dx, 1;                    //dx = dx OR 1
	mov xdir, dx;                 //xdir = sgn(vx)
      }

      //incy = abs(vy);
      //ydir = sgn(vy);
      asm{
	mov ax, vy;                   //ax = vy
	cwd;                          //if vy < 0 then dx = -1 else dx = 0
	xor ax, dx;                   //ax = vy XOR dx
	//sub ax, dx;                   //ax = (vy XOR dx) - dx
	mov incy, ax;                 //incy = abs(vy)
	or  dx, 1;                    //dx = dx OR 1
	mov ydir, dx;                 //ydir = sgn(vy)
      }

      xtemp = xtempinit;
      if (xdir > 0) { //this is the same as checking (vx > 0)
	xtemp = 64 - xtemp;
      }
      ytemp = ytempinit;
      if (ydir > 0) { //this is the same as checking (vy > 0)
	ytemp = 64 - ytemp;
      }

      //int d = (incy >> 6) * xtemp - (incx >> 6) * ytemp;
      asm {
	mov ax, incy;                 //ax = incy
	mov cl, 6;                    //cl = 6
	sar ax, cl;                   //ax = incy >> 6
	imul xtemp;                   //dx:ax = (incy >> 6) * xtemp
	mov di, ax;                   //di = (incy >> 6) * xtemp
	mov ax, incx;                 //ax = incx
	sar ax, cl;                   //ax = incx >> 6
	imul ytemp;                   //dx:ax = (incx >> 6) * ytemp
	sub di, ax;                   //di = (incy>>6)*xtemp - (incx>>6)*ytemp
      }
lDDA:
      //while (-1) {
      //  if (d < 0) {
	asm {
	  or di, di                   //if di >= 0
	  jge ldPositive;             //  then goto ldPositive
	}
	  //Dark side
	  //xscan += xdir;
	  asm {
	    mov dx, xdir;             //dx = xdir
	    add xscan, dx;            //xscan = xscan + xdir
	  }
	  //if (board[xscan+yscan*64]) {
	  asm {
	    mov bx, yscan;            //bx = yscan
	    shl bx, cl;               //yscan = yscan << 6
	    add bx, xscan;            //bx = (yscan << 6) + xscan
	    cmp byte ptr board[bx], 0;//if board[(yscan << 6) + xscan] = 0
	    je  lNoWallDarkSide;      //  then goto lNoWallDarkSide
	  }
	    //hx = xscan * 64;
	    //if (xdir < 0) { //this is the same as checking (vx < 0)
	    //  hx += 64;
	    //}

	    //hx = xscan * 64 + (xdir & 64);
	    asm {
	      mov ax, xscan;          //ax = xscan
	      shl ax, cl;             //ax = xscan << 6
	      and dx, 64;             //dx = xdir AND 64
	      add ax, dx;             //ax = (xscan << 6) + (xdir AND 64)
	      mov hx, ax;             //hx = (xscan << 6) + (xdir AND 64)
	    }

	    //hy = playerposy + (vy >> 8) * (hx - playerposx) / (vx >> 8);
	    asm {
	      mov ax, vy;             //ax = vy
	      mov cl, 8;              //cl = 8
	      sar ax, cl;             //ax = vy >> 8
	      mov dx, hx;             //dx = hx
	      sub dx, playerposx;     //dx = hx - playerposx
	      imul dx;                //dx:ax = (vy >> 8) * (hx - playerposx)
	      cwd;                    //dx = (ax < 0x8000 ? 0 : 0xFFFF)
	      mov bx, vx;             //bx = vx
	      sar bx, cl;             //bx = vx >> 8
	      idiv bx;                //ax = dx:ax / bx, dx = dx:ax mod bx
	      mov dx, playerposy;     //dx = playerposy
	      add dx, ax;             //dx = playerposy + ax
	      mov hy, dx;             //hy = playerposy + ax
	    }

	    //texx = hy & 63;
	    asm {
	      mov ax, hy;             //ax = hy
	      and ax, 63;             //ax = hy AND 63
	      mov texx, ax;           //texx = hy AND 63
	    }
	    //This is only needed for texture mapping.
	    //It can be removed, because you won't notice
	    //it at such a low resolution.
	    //if (xdir < 0) { //this is the same as checking (vx < 0)
	    //  texx = 63 - texx;
	    //}
	    //break;
	    asm {
	      jmp lComputeDistance;   //goto lComputeDistance
	    }
lNoWallDarkSide:
	  //}
	  //d += incy;
	  asm {
	    add di, incy;             //d = d + incy
	    jmp lDDA;                 //goto lDDA
	  }
	//}
ldPositive:
	//else {
	  //Light side
	  //yscan += ydir;
	  asm {
	    mov dx, ydir;             //dx = ydir
	    add yscan, dx;            //yscan = yscan + ydir
	  }
	  //if (board[xscan+yscan*64]) {
	  asm {
	    mov bx, yscan;            //bx = yscan
	    shl bx, cl;               //bx = yscan << 6
	    mov ax, bx;               //ax = yscan << 6
	    add bx, xscan;            //bx = (yscan << 6) + xscan
	    cmp byte ptr board[bx], 0;//if board[(yscan << 6) + xscan] = 0
	    je  lNoWallLightSide;     //  then goto lNoWallLightSide
	  }
	    //hy = yscan * 64;
	    //if (ydir < 0) { //this is the same as checking (vy < 0)
	    //   hy += 64;
	    //}
	    //hy = yscan * 64 + (ydir & 64);
	    asm {
	      and dx, 64;             //dx = ydir AND 64
	      add ax, dx;             //ax = (yscan << 6) + (ydir AND 64)
	      mov hy, ax;             //hy = (yscan << 6) + (ydir AND 64)
	    }

	    //hx = playerposx + (vx >> 8) * (hy - playerposy) / (vy >> 8);
	    asm {
	      mov ax, vx;             //ax = vx
	      mov cl, 8;              //cl = 8
	      sar ax, cl;             //ax = vx >> 8
	      mov dx, hy;             //dx = hy
	      sub dx, playerposy;     //dx = hy - playerposy
	      imul dx;                //dx:ax = (vx >> 8) * (hy - playerposy)
	      cwd                     //dx = (ax < 0x8000 ? 0 : 0xFFFF)
	      mov bx, vy;             //bx = vy
	      sar bx, cl;             //bx = vy >> 8
	      idiv bx;                //ax = dx:ax / bx, dx = dx:ax mod bx
	      mov dx, playerposx;     //dx = playerposx
	      add dx, ax;             //dx = playerposx + ax
	      mov hx, dx;             //hx = playerposx + ax
	    }

	    //texx = hx & 63;
	    asm {
	      mov ax, hx;             //ax = hx
	      and ax, 63;             //ax = hx AND 63
	      mov texx, ax;           //texx = hx AND 63
	    }
	    //This is only needed for texture mapping.
	    //It can be removed, because you won't notice
	    //it at such a low resolution.
	    //if (ydir > 0) { // this is the same as checking (vy > 0)
	    //  texx = 63 - texx;
	    //}
	    //break;
	    asm {
	      jmp lComputeDistance;   //goto lComputeDistance
	    }
lNoWallLightSide:
	  //}
	  //d -= incx;
	  asm {
	    sub di, incx;             //d = d - incx;
	    jmp lDDA;                 //goto lDDA
	  }
	//}
      //}
lComputeDistance:

      //int dist = (((cosang >> 11) * (hx - playerposx)) >> 1) +
      //           (((sinang >> 11) * (hy - playerposy)) >> 1);
      asm {
	mov ax, cosang;               //ax = cosang
	mov cl, 11;                   //cl = 11
	sar ax, cl;                   //ax = cosang >> 11
	mov dx, hx;                   //dx = hx
	sub dx, playerposx;           //dx = hx - playerposx
	imul dx;                      //dx:ax = (cosang>>11) * (hx-playerposx)
	sar ax, 1;                    //ax = ((cosang>>11)*(hx-playerposx))>>1
	mov di, ax;                   //dist=((cosang>>11)*(hx-playerposx))>>1

	mov ax, sinang;               //ax = sinang
	sar ax, cl;                   //ax = sinang >> 11
	mov dx, hy;                   //dx = hy
	sub dx, playerposy;           //dx = hy - playerposy
	imul dx;                      //dx:ax = (sinang>>11) * (hy-playerposy)
	sar ax, 1;                    //ax = ((sinang>>11)*(hy-playerposy))>>1
	add di, ax;                   //dist+=(sinang>>11)*(hy-playerposy)>>1
      }

      //Find the ceiling and floor borders
      //if (dist == 0) {
      asm {
	or di, di;                    //if dist != 0
	jne lDistNotZero;             //  then goto lDistNotZero
      }
	//sy1 = 0;
	//sy2 = YDIM;
	asm {
	  mov sy1, 0;                 //sy1 = 0
	  mov sy2, YDIM;              //sy2 = YDIM
	  jmp lComputeTexincy;        //goto lComputeTexincy
	}
      //}
lDistNotZero:
      //else {
	//sy1 = YDIM / 2 - HVAL * 128 / dist;
	//sy2 = YDIM / 2 + HVAL * 128 / dist;
	asm {
	  mov ax, HVAL * 128;         //ax = HVAL * 128
	  cwd;                        //dx = 0
	  idiv di;                    //ax = dx:ax / dist, dx = dx:ax mod dist
	  mov dx, YDIM / 2;           //dx = YDIM / 2
	  sub dx, ax;                 //dx = YDIM / 2 - HVAL * 128 / dist
	  mov sy1, dx;                //sy1 = YDIM / 2 - HVAL * 128 / dist

	  add ax, YDIM / 2;           //ax = HVAL * 128 / dist + YDIM / 2
	  mov sy2, ax;                //sy2 = HVAL * 128 / dist + YDIM / 2
	}
	//if (sy1 < 0)    {
	//  sy1 = 0;
	//}
	asm {
	  cmp sy1, 0;                 //if sy1 >= 0
	  jge lsy1IsSet;              //  then goto lsy1IsSet
	  mov sy1, 0;                 //sy1 = 0
	}
lsy1IsSet:
	//if (sy2 > YDIM) {
	//  sy2 = YDIM;
	//}
	asm {
	  cmp sy2, YDIM;              //if sy2 <= YDIM
	  jle lComputeTexincy;        //  then goto lComputeTexincy
	  mov sy2, YDIM;              //sy2 = YDIM
	}
      //}

lComputeTexincy:
      //This is wrong when the wall fills the whole column.
      //But it's faster than using the right way.
      //texincy = (64 << 8) / (sy2 - sy1 + 1); // + 1 to eliminate division by 0
      asm {
	mov bx, sy2;                  //bx = sy2
	sub bx, sy1;                  //bx = sy2 - sy1
	inc bx;                       //bx = sy2 - sy1 + 1
	mov ax, 64 * 256;             //ax = 64 << 8
	cwd;                          //dx = 0
	idiv bx;                      //ax = dx:ax / bx, dx = dx:ax mod bx
	mov texincy, ax;              //texincy = (64 << 8) / (sy2 - sy1 + 1)
      }

      //sy1 *= BPL;
      //sy2 *= BPL;
      asm {
	mov cx, BPL;                  //cx = BPL
	mov ax, sy1;                  //ax = sy1
	imul cx;                      //dx:ax = sy1 * BPL
	mov sy1, ax;                  //sy1 = sy1 * BPL
	mov ax, sy2;                  //ax = sy2
	imul cx;                      //dx:ax = sy2 * BPL
	mov sy2, ax;                  //sy2 = sy2 * BPL
      }

      //Draw vertical line at column column
      asm {
	mov ax, 0xB800;               //es = 0xB800
	mov es, ax;                   //

	mov si, column;               //offset = column

	jmp lCeilingCheck;            //goto lCeilingCheck
      }
lDrawCeiling:
      asm {
	mov byte ptr es:[si], 8;      //pokeb(0xB800, offset, 8)
	add si, BPL;                  //offset = offset + BPL
      }
lCeilingCheck:
      asm {
	cmp si, sy1;                  //if offset < sy1 then //(offset ~= y)
	jl  lDrawCeiling;             //  then goto lDrawCeiling
      }

      //texoffset = (texx + (board[xscan + yscan * 64] - 1) * 64) * 64;
      asm {
	xor dx, dx;                   //texy = 0
	mov cl, 6;                    //cl = 6
	mov ax, yscan;                //ax = yscan
	shl ax, cl;                   //ax = ax * 64
				      //ax == yscan * 64
	mov bx, xscan;                //bx = xscan
	add bx, ax;                   //bx = bx + ax
	mov al, board[bx];            //al = board[bx]
	cbw;                          //ah = 0
	dec ax;                       //ax = ax - 1
	shl ax, cl;                   //ax = ax * 64
	add ax, texx;                 //ax = ax + texx
	shl ax, cl;                   //ax = ax * 64
	mov cx, ax;                   //texoffset = ax

				      //cx == texoffset
				      //dx == texy

	jmp lWallCheck                //goto lWallCheck
      }

lDrawWall:
      asm {
	mov bl, dh;                   //bl = high byte of texy
	xor bh, bh;                   //bh = 0
	add bx, cx;                   //bx = bx + texoffset
				      //bx == (texy >> 8) + texoffset
	mov al, byte ptr texture[bx]; //al = texture[bx]
	mov byte ptr es:[si], al;     //pokeb(0xB800, offset, al)
	add si, BPL;                  //offset = offset + BPL
	add dx, texincy;              //texy = texy + texincy
      }
lWallCheck:
      asm {
	cmp si, sy2;                  //if offset < sy2 //(offset ~= y)
	jl  lDrawWall;                //  then goto lDrawWall
	jmp lFloorCheck;              //  else goto lFloorCheck
      }

lDrawFloor:
      asm {
	mov byte ptr es:[si], 7+7*16; //pokeb(0xB800, offset, 7 + 7 * 16);
	add si, BPL;                  //offset = offset + BPL
      }
lFloorCheck:
      asm {
	cmp si, YDIM * BPL;           //if offset < YDIM //(offset ~= y)
	jl  lDrawFloor;               //  then goto lDrawFloor
      }

      vx -= sinang / (XDIM / 2);
      vy += cosang / (XDIM / 2);
    }
#ifndef DEMORUN
    frames++;

    //Keyboard control code begin
    if (kbhit()) {
      int c;
      int extended = 0;
      c = getch();
      if (!c) {
	extended = getch();
      }

      if (c == 27) {
        quit = -1;
      }
      else if (!c && (extended == 75)) {
	playerang -= 4;
      }
      else if (!c && (extended == 77)) {
	playerang += 4;
      }
      else if ((c == '<') || (c == ',')) {
	playerposx += sinang >> 10;
	playerposy -= cosang >> 10;
      }
      else if ((c == '>') || (c == '.')) {
	playerposx -= sinang >> 10;
	playerposy += cosang >> 10;
      }
      else if (!c && (extended == 72)) {
	playerposx += cosang >> 10;
	playerposy += sinang >> 10;
      }
      else if (!c && (extended == 80)) {
	playerposx -= cosang >> 10;
	playerposy -= sinang >> 10;
      }
    }
    //Keyboard control code end
#endif
  }

  end = clock();

#ifdef DEMORUN
  waitForKeyPress();
  showStatistics((end-start)/CLK_TCK, 614/((end-start)/CLK_TCK));
#else
  showStatistics((end-start)/CLK_TCK, frames/((end-start)/CLK_TCK));
#endif

  return 0;
}